/*
tlib - Unicode String Class
*/

#include <ctype.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <wchar.h>
#pragma hdrstop

#include "tlib.h"

using namespace tlib;

#if defined(_MSC_VER)
	#define vsnwprintf _vsnwprintf
#elif defined(__BORLANDC__)
	#define wmemcmp(s1, s2, size) memcmp(s1, s2, (size) * sizeof(wchar_t))
#endif

//---------------------------------------------------------------------------



TString::TString()
{
	FReferenceCount = NULL;
	FData = NULL;
	FSize = 0;
}



TString::TString(int size, const wchar_t* data)
{
	if (size != 0 && data == NULL)
		throw TException(L"Invalid arguments");

	if (size == 0)
	{
		FReferenceCount = NULL;
		FData = NULL;
		FSize = 0;
	}
	else
	{
		FReferenceCount = new int;
		*FReferenceCount = 1;
		FSize = size;
		FData = new wchar_t [FSize + 1];
		wmemcpy(FData, data, FSize);
		FData[FSize] = L'\0';
	}
}



TString::TString(int c, int count)
{
	if (c < 0 || c >= 0x110000)
		throw TException(L"Invalid character");
	if (count < 0)
		throw TException(L"Invalid count");

	if (count == 0)
	{
		FReferenceCount = NULL;
		FData = NULL;
		FSize = 0;
		return;
	}
	FReferenceCount = new int;
	*FReferenceCount = 1;
	if (c < 0xd800 || (c >= 0xe000 && c < 0x10000))
	{
		FSize = count;
		FData = new wchar_t [FSize + 1];
		for (int i = 0; i < count; i++)
			FData[i] = c;
	}
	else
	{
		FSize = count * 2;
		FData = new wchar_t [FSize + 1];
		wchar_t e0 = 0xd800 | (c >> 10 & 0x3ff);
		wchar_t e1 = 0xdc00 | (c & 0x3ff);
		for (int i = 0; i < FSize;)
		{
			FData[i++] = e0;
			FData[i++] = e1;
		}
	}
	FData[FSize] = L'\0';
}



TString::TString(const TString& str)
{
	FReferenceCount = str.FReferenceCount;
	if (FReferenceCount)
		(*FReferenceCount)++;
	FData = str.FData;
	FSize = str.FSize;
}



TString::~TString()
{
	Release();
}



TString& TString::operator =(const TString& str)
{
	Release();
	FReferenceCount = str.FReferenceCount;
	if (FReferenceCount)
		(*FReferenceCount)++;
	FData = str.FData;
	FSize = str.FSize;
	return *this;
}



TString TString::operator +(const TString& str) const
{
	if (str.FSize == 0)
		return *this;
	if (FSize == 0)
		return str;

	TString r;
	r.FReferenceCount = new int;
	*r.FReferenceCount = 1;
	r.FSize = FSize + str.FSize;
	r.FData = new wchar_t [r.FSize + 1];
	wmemcpy(r.FData, FData, FSize);
	wmemcpy(r.FData + FSize, str.FData, str.FSize);
	r.FData[r.FSize] = L'\0';
	return r;
}



TString& TString::operator +=(const TString& str)
{
	if (!FData)   // string is empty, just assign to str
	{
		Release();
		FReferenceCount = str.FReferenceCount;
		if (FReferenceCount)
			(*FReferenceCount)++;
		FData = str.FData;
		FSize = str.FSize;
	}
	else if (str.FSize != 0)
	{
		int Size = FSize + str.FSize;
		wchar_t *Data = new wchar_t [Size + 1];
		wmemcpy(Data, FData, FSize);
		wmemcpy(Data + FSize, str.FData, str.FSize);
		Data[Size] = L'\0';

		Release();
		FReferenceCount = new int;
		*FReferenceCount = 1;
		FData = Data;
		FSize = Size;
	}
	return *this;
}



bool TString::operator ==(const TString& str) const
{
	if (FSize != str.FSize)
		return false;
	if (FSize == 0)
		return true;   // because Sizes are equal
	return !wmemcmp(FData, str.FData, FSize);
}



bool TString::operator !=(const TString& str) const
{
	if (FSize != str.FSize)
		return true;
	if (FSize == 0)
		return false;   // because Sizes are equal
	return wmemcmp(FData, str.FData, FSize);
}



int TString::_GetLength() const
{
	int Length = Size;
	for (int i = 0; i < Size; i++)
	{
		wchar_t e = FData[i];
		if (e >= 0xdc00 && e < 0xe000)
			throw TException(L"Invalid string data");
		if (e >= 0xd800 && e < 0xdc00)
		{
			i++;
			if (i == Size)
				throw TException(L"Invalid string data");
			wchar_t e1 = FData[i];
			if (e1 < 0xdc00 && e1 >= 0xe000)
				throw TException(L"Invalid string data");
			Length--;
		}
	}
	return Length;
}



int TString::operator [](int index) const
{
	int Pos = GetPos(index);
	if (Pos == FSize)
		throw TException(L"Invalid index");
	wchar_t e = FData[Pos];
	if (e >= 0xdc00 && e < 0xe000)
		throw TException(L"Invalid string data");
	if (e >= 0xd800 && e < 0xdc00)
	{
		Pos++;
		if (Pos >= FSize)
			throw TException(L"Invalid string data");
		wchar_t e1 = FData[Pos];
		if (e1 < 0xdc00 || e1 >= 0xe000)
			throw TException(L"Invalid string data");
		return ((e & 0x3ff) << 10l | (e1 & 0x3ff)) + 0x10000;
	}
	return (int)e;
}



/**
Returns a substring from the specified index to the end of the string.
*/
TString TString::Substring(int index) const
{
	int Start = GetPos(index);
	return TString(FSize - Start, FData + Start);
}



/**
Returns the specified range of characters as a string.
*/
TString TString::Substring(int index, int length) const
{
	if (index < 0 || length < 0)
		throw TException(L"Invalid arguments");

	int Start = GetPos(index);
	int Size = GetSize(Start, length);
	if (Start + Size > FSize)
		throw TException(L"Substring length cannot exceed string length");
	return TString(Size, FData + Start);
}



/**
Returns the character index of the first occurence of substring 'str' after index 'start'
in the string. Returns -1 if the substring was not found.
*/
int TString::FindSubstring(TString str, int start) const
{
	int s = str.FSize;
	int p = 0;

	if (!s)
		throw TException(L"Cannot match empty string");

	int StartIndex;
	int Index = start;
	for (int i = GetPos(start); i < FSize; i++)
	{
		if (FData[i] == str.FData[p])
		{
			if (p == 0)
				StartIndex = Index;
			if (++p == s)
				return StartIndex;
		}
		else
			p = 0;
		if (!(FData[i] >= 0xdc00 && FData[i] < 0xe000))
			Index++;
	}
	return -1;
}



/**
Returns the index of the first occurence of character 'c' after 'startIndex'.
Returns -1 if the character is not in the string.
*/
int TString::FindChar(int c, int startIndex) const
{
	int Pos = GetPos(startIndex);
	for (int i = startIndex; ; i++)
	{
		if (Pos == FSize)
			return -1;
		wchar_t e = FData[Pos++];
		if (e >= 0xdc00 && e < 0xe000)
			throw TException(L"Invalid string data");
		if (e < 0xd800 || e >= 0xdc00)
		{
			if ((int)e == c)
				return i;
		}
		else
		{
			if (Pos == FSize)
				throw TException(L"Invalid string data");
			wchar_t e1 = FData[Pos++];
			int d = ((e & 0x3ff) << 10l | (e1 & 0x3ff)) + 0x10000;
			if (d == c)
				return i;
		}
	}
}



/**
Returns the index of the last occurence of character 'c' in the string.
Returns -1 if the character is not in the string.
*/
int TString::FindLastChar(int c) const
{
	int Index = -1;
	int Pos = 0;
	for (int i = 0; ; i++)
	{
		if (Pos == FSize)
			return Index;
		wchar_t e = FData[Pos++];
		if (e >= 0xdc00 && e < 0xe000)
			throw TException(L"Invalid string data");
		if (e < 0xd800 || e >= 0xdc00)
		{
			if ((int)e == c)
				Index = i;
		}
		else
		{
			if (Pos == FSize)
				throw TException(L"Invalid string data");
			wchar_t e1 = FData[Pos++];
			int d = ((e & 0x3ff) << 10l | (e1 & 0x3ff)) + 0x10000;
			if (d == c)
				Index = i;
		}
	}
}



/**
Returns the buffer index of the character at the specified index.
*/
int TString::GetPos(int index) const
{
	if (index < 0)
		throw TException(L"Index out of range");
	int Pos = 0;
	for (int i = 0; 1; i++)
	{
		if (i == index)
			return Pos;
		if (Pos >= FSize)
			throw TException(L"Index out of range");
		wchar_t e = FData[Pos++];
		if (e >= 0xdc00 && e < 0xe000)
			throw TException(L"Invalid string data");
		if (e >= 0xd800 && e < 0xdc00)
		{
			if (Pos >= FSize)
				throw TException(L"Invalid string data");
			wchar_t e1 = FData[Pos++];
			if (e1 < 0xdc00 || e1 >= 0xe000)
				throw TException(L"Invalid string data");
		}
	}
}



/**
Returns the number of elements in the specified range of text.
*/
int TString::GetSize(int startPos, int length) const
{
	if (startPos > FSize)
		throw TException(L"Invalid start pos");

	int Pos = startPos;
	for (int i = 0; ; i++)
	{
		if (i == length)
			return Pos - startPos;
		if (Pos == FSize)
			throw TException(L"Invalid length");
		wchar_t e = FData[Pos++];
		if (e >= 0xdc00 && e < 0xe000)
			throw TException(L"Invalid string data");
		if (e >= 0xd800 && e < 0xdc00)
		{
			if (Pos == FSize)
				throw TException(L"Invalid string data");
			wchar_t e1 = FData[Pos++];
			if (e1 < 0xdc00 || e1 >= 0xe000)
				throw TException(L"Invalid string data");
		}
	}
}



int TString::Iterator(int i) const
{
	return GetPos(i);
}



void TString::Next(int& p) const
{
	wchar_t e = FData[p++];
	if (e >= 0xdc00 && e < 0xe000)
		throw TException(L"Invalid string data");
	if (e >= 0xd800 && e < 0xe000)
	{
		wchar_t e1 = FData[p++];
		if (e1 < 0xdc00 || e1 >= 0xe000)
			throw TException(L"Invalid string data");
	}
}



int TString::Last() const
{
	if (FSize == 0)
		return -1;
	int Pos = FSize - 1;
	wchar_t e = FData[Pos];
	if (e >= 0xdc00 && e < 0xe000)
	{
		if (Pos == 0)
			throw TException(L"Invalid string data");
		--Pos;
		e = FData[Pos];
		if (!(e >= 0xd800 && e < 0xdc00))
			throw TException(L"Invalid string data");
	}
	return Pos;
}



void TString::Prev(int& p) const
{
	--p;
	if (p == -1)
		return;
	wchar_t e = FData[p];
	if (e >= 0xdc00 && e < 0xe000)
	{
		if (p == 0)
			throw TException(L"Invalid string data");
		--p;
		e = FData[p];
		if (!(e >= 0xd800 && e < 0xdc00))
			throw TException(L"Invalid string data");
	}
}



int TString::GetChar(int p) const
{
	if (p < 0 || p >= FSize)
		throw TException(L"Invalid iterator");
	wchar_t e = FData[p];
	if (e >= 0xdc00 && e < 0xe000)
		throw TException(L"Invalid string data");
	if (e < 0xd800 || e >= 0xe000)
		return (int)e;
	if (p + 1 == FSize)
		throw TException(L"Invalid string data");
	wchar_t e1 = FData[p + 1];
	if (e1 < 0xdc00 || e1 >= 0xe000)
		throw TException(L"Invalid string data");
	return ((e & 0x3ff) << 10l | (e1 & 0x3ff)) + 0x10000;
}



TString TString::SPrintF(const wchar_t* fmt, ...)
{
	va_list arg;
	va_start(arg, fmt);
	int Length = vsnwprintf(NULL, 0, fmt, arg);
	wchar_t *Data = new wchar_t [Length + 1];
	vsnwprintf(Data, Length + 1, fmt, arg);
	TString r(Length, Data);
	delete [] Data;
	return r;
}



TString TString::IntToStr(int n)
{
	wchar_t Str[33];
	_itow(n, Str, 10);
	return TString(wcslen(Str), Str);
}



//---------------------------------------------------------------------------



bool TChar::IsAlpha(int c)
{
	if (c < 0 || (c >= 0xd800 && c < 0xe000) || c >= 0x110000)
		throw TException(L"Invalid character code");
	if (c >= 0x10000)
		return false;
	return (bool)iswalpha(c);
}



bool TChar::IsAlphaNumeric(int c)
{
	if (c < 0 || (c >= 0xd800 && c < 0xe000) || c >= 0x110000)
		throw TException(L"Invalid character code");
	if (c >= 0x10000)
		return false;
	return (bool)iswalnum(c);
}



bool TChar::IsWhiteSpace(int c)
{
	if (c < 0 || (c >= 0xd800 && c < 0xe000) || c >= 0x110000)
		throw TException(L"Invalid character code");
	if (c >= 0x10000)
		return false;
	return (bool)iswspace(c);
}